/*
 * Function.cpp
 *
 *  Created on: Oct 9, 2017
 *      Author: julien
 */

#include "Function.hpp"

using namespace cv;
using namespace std;

Function::Function(int Lookupsize) : Lookupsize(Lookupsize + 1){

	// make just enough space for all the need factorial numbers
	fak = vector<double>(this->Lookupsize);

	// make just enough space for all the binomial coefficients
	bin_co = vector<vector<double>>(this->Lookupsize);

	// the lookup table works like so : fak(n) = n!
	// for calculating the factorial numbers we start with 0! = 1
	fak.at(0) = 1;
	// and then just multiply the number before that by the current index
	for(int i = 1; i < this->Lookupsize; i++){
		fak.at(i) = fak.at(i-1)*i;
	}

	// the binomial coefficient will work like so : bin_co [n][k] = (n)
	//																(k)
	// look at the wikipedia entry to see how this is calculated
	for(int n = 0; n < this->Lookupsize; n++){

		bin_co.at(n) = vector<double>(n+1);

		for(int k = 0; k <= n; k++){
			bin_co.at(n).at(k) = fak[n]/(fak[k] * fak[n-k]);
		}
	}

	// now create enough space for a lookup table for both sinus and cosines
	// 3600, so we can calculate angles up to one decimal number
	si = vector<double>(36000);
	co = vector<double>(36000);
	ta = vector<double>(9000);
	asi = vector<double>(2001);
	aco = vector<double>(2001);
	ata = vector<double>(10001);


	// use a conversion from radian to degree to make it more user friendly
	for(int i = 0; i < 36000; i++){
		si[i] = sin(M_PI * i/18000);
	}
	// same goes for cosinus
	for(int i = 0 ; i < 36000; i++){
		co[i] = cos(M_PI * i/18000);
	}

	// use a conversion from radian to degree to make it more user friendly
	for(int i = 0; i < 2001; i++){
		asi[i] = 180 * asin((i/1000.0) - 1) / M_PI;
	}
	// same goes for cosinus
	for(int i = 0 ; i < 2001; i++){
		aco[i] = 180 * acos((i/1000.0) - 1) / M_PI;
	}

	for(int i = 0; i < 9000; i++){
		ta[i] = tan(M_PI * i/18000);
	}
	for(int i = 0 ; i < 10001; i++){
		ata[i] = (180 * (atan(i/1000.0)) / M_PI);
	}

}


double Function::sinus(double angle){

	// find the corresponding index in the lookup table to the angle
	// smiply find the corresponding index by multiplying by 10
	// and then use the identity of the angles inside circles
	int a = ((int)(angle * 100))%36000;

	// sin(-x) = -sin(x)
	if(a > 0){
		return si[a];
	}
	else if(a < 0){
		return si[-a] * (-1);
	}
	return 0;
}

double Function::cosinus(double angle){

	// find the corresponding index in the lookup table to the angle
	// smiply find the corresponding index by multiplying by 10
	// and then use the identity of the angles inside circles
	int a = ((int)(angle * 100))%36000;

	// cos(-x) = cos(x)
	if(a > 0){
		return co[a];
	}
	else if(a < 0){
		return co[-a];
	}
	return 1;
}


double Function::tangens(double angle){

	// if we are outside the mathematically defined range of tangens,
	if((((int)angle) + 90)%180 == 0){
		// then throw the appropriate exception
		throw math_range_error();
	}

	// find the corresponding index in the lookup table to the angle
	// smiply find the corresponding index by multiplying by 10
	// and then use the identity of the angles inside circles
	int a = ((int)(angle * 100))%36000;

	// tan(-x) = -tan(x)
	if(a > 0){
		return ta[a];
	}
	else if(a < 0){
		return -ta[-a];
	}
	return 0;
}


double Function::arc_sinus(double fracture){

	// if we are outside the mathematically defined range of arcsin,
	if(fracture > 1 || fracture < -1){
		// then throw the appropriate exception
		throw math_range_error();
	}

	// calculate the corresponding index for the lookup table
	int x = (int)((fracture + 1)* 1000);

	return asi[x];

}


double Function::arc_cosinus(double fracture){

	// if we are outside the mathematically defined range of arcsin,
	if(fracture > 1 || fracture < -1){
		// then throw the appropriate exception
		throw math_range_error();
	}

	// calculate the corresponding index for the lookup table
	int x = (int)((fracture + 1)* 1000);

	return aco[x];

}

double Function::arc_tangens(double fracture){

	// calculate the corresponding index for the lookup table
	if(fracture > 100) fracture = 10;
	else if(fracture < -10)	fracture = -10;

	int x = (int)(fracture* 1000);

	if(x >= 0){
		return ata[x];
	}
	else{
		return -ata[-x];
	};
}


vector<double> Function::square_fit(vector<Function::pPoint> points, int grad){


	// for further understanding what is meant inside these comments look at the documentation,
	// or look up on the least square fit algorithm


	// number of parameters
	int num_param = grad + 1;

	// number of powers needed for the calculation
	int num_powers = grad*2 + 1;

	//produce an empty mat in the size of grad + 1
	Mat M(Size(num_param, num_param), CV_64FC1);
	Mat invM(Size(num_param, num_param), CV_64FC1);

	// a vector for storing the values of all the sums of the final matrix A
	// where the index of the vector will determine the summed power as:
	// [i] -> sum(x^i)
	vector<double> sums_of_xi_powers(num_powers);

	// a vector for storing the final parameters of the function
	vector<double> ret(num_param);

	// a vector for storing all the sums of yi for the final calculation
	vector<double> sums_of_yi(num_param);

	// a vector for temporarily storing all the powers of xi during a for cycle
	// where the index of the vector will determine the power n by n = i
	vector<double> powers_of_xi(num_powers);

	// declaration of the temporary values outside of the loop
	double xi;
	double yi;
	int size = points.size();

	// prepare the first index of the vectors as seen in the documentation
	// (least square fit)
	powers_of_xi[0] = 1;
	sums_of_xi_powers[0] = size;


	// for all the points inside the given points
	for(int i = 0; i < size; i++){

		xi = points.at(i).x;
		yi = points.at(i).y;

		// for all the powers
		for(int n = 1; n < num_powers; n++){
			// calculate all the powers one after another
			powers_of_xi[n] = powers_of_xi[n-1] * xi;

			// add the powers of the xis to the corresponding sum
			sums_of_xi_powers[n] += powers_of_xi[n];
		}

		// calculate the yis one after another
		for(int n = 0; n < num_param; n++){
			sums_of_yi[n] += yi * powers_of_xi[n];
		}

	}

	double* row_pointer;

	// next populate the matrix A with the calculated values
	for(int row = 0; row < num_param; row++){

		row_pointer = M.ptr<double>(row);

		for(int col = 0; col < num_param; col++){

			// Mathematisch : M(i,j) = sum( xi^( i+j ) )
			row_pointer[col] = sums_of_xi_powers[ row + col ];

		}
	}

	// invert the matrix
	invM = M.inv(DECOMP_CHOLESKY);

	// calculate the parameters of the function
	// by multiplication of the vector y and the inverse Matrix of M
	double param_i;
	for(int i = 0; i < num_param ; i++){
		param_i = 0;

		row_pointer = invM.ptr<double>(i);

		for(int col = 0; col < num_param; col++){
			param_i += row_pointer[col] * sums_of_yi.at(col);
		}

		ret.at(i) = param_i;
	}

	return ret;

}


void Function::vec_fit(vector<pPoint>* points, Function::vec_func* out, int grad){

	// first we determine the sample size
	int size = points->size();

	// now the vectors for storing the value pairs are created
	// the basic premise is the usage of least-square-fit independently for x and y
	// this is also the reason why the points must(!) be ordered
	vector<pPoint> indexed_x_points(size);
	vector<pPoint> indexed_y_points(size);

	// now populate the vectors with the values, where the pair is determined via their index
	for(int i = 0; i < size; i++){
		indexed_x_points.at(i) = pPoint(i, points->at(i).x);
		indexed_y_points.at(i) = pPoint(i, points->at(i).y);
	}

	// now fit both polynomial functions independently
	out->poly_x = Function::square_fit(indexed_x_points, grad);
	out->poly_y = Function::square_fit(indexed_y_points, grad);
	// we always start at 0
	out->start_t = 0;
	// and the end_t is determined over the sample size
	out->end_t = size - 1;

}

void Function::move_vec_func(Function::vec_func* v, double delta_x, double delta_y){

	// to move a vector function you simply need to add the delta-value
	// to the Constant inside that polynomial function
	v->poly_x.at(0) += delta_x;
	v->poly_y.at(0) += delta_y;

}


vector<double> Function::differentiate(vector<double>* poly){

	// this short block is for error handling (see later as to why)
	// if the size isn't big enough, then just return a constant zero over and over
	int size = poly->size();
	if(size == 1){
		return {0};
	}

	// create a temporary storrage for the output
	vector<double> temp;

	// since this is a standardized polynomial function we deal with,
	// one only needs to be concerned with the sum rule and the power rule
	// (d/dx)( a * x ^ n ) = n*a*x^(n-1)
	// we start at one since the constant value always vanishes in this version
	// differentiation.
	// this is also the reason, why the error handling needed to be added, so this function
	// doesn't return empty vectors
	for(int i = 1; i < size; i++){
		temp.push_back(poly->at(i) * i);
	}
	return temp;
}


double Function::evaluate(vector<double>* poly, double x){

	// determine of many parameter the polynomial function has
	int params = poly->size();
	// to calculate the powers of x in O(n) we calculate it with this helper variable
	double power_of_x = 1;

	// this will be the final output value
	// first we "add" the constant of the function
	// remember that the power of x corresponding to the parameter is determined via the index
	double y = poly->at(0);


	for(int p = 1; p < params; p++){

		// calculate the next power
		power_of_x *= x;

		// the output y is p1*x + p2*x² + p3*x³ + ...
		y += power_of_x * poly->at( p );

	}

	return y;
}


vector<double> Function::move_y(vector<double> poly, double y){

	// just add the delta value to the constant to move it in the y-direction
	// this only works for polynomial functions
	poly.at(0) = poly.at(0) + y;
	return poly;
}


vector<double> Function::move_x(vector<double> poly, double x){

	// a simple catch, to not calculate the whole function
	// since the delta value might be zero,
	// since most inputs are automated based upon the ego motion
	if(x == 0){
		return poly;
	}

	// now start with the real function

	// calculate size once, since it is important for the rest of the function
	int size = poly.size();

	//clacuate the powers of x
	// to calculate the powers of x in O(n) we calculate it with this helper variable
	// since this algorithm needs it multiple times, they are stored separately
	// based upon their index
	vector<double> powers_of_x(size);
	powers_of_x[0] = 1;

	// powers are stored based upon their index, which means {x⁰, x¹, x³, ..., x^n}
	for(int i = 1; i < size; i++){
		powers_of_x[i] = powers_of_x[i-1] * x;
	}

	// initialize the temporary vector for the output
	vector<double> ret(size);

	// param_i
	double param_i = 0; // the parameter the will be at index i in the ouput vector
	int sign; // sign = vorzeichen

	for(int i = 0; i < size; i++){

		// reset the parameter and the sign
		sign = -1;
		param_i = 0;

		for(int k = i; k < size; k++){

			sign *= -1;

			// to understand the formula read the documentation
			// TODO documentation
			param_i +=  bin_co[k][i] * poly.at(k) * sign * powers_of_x[k-i];

		}

		// insert the calculated parameter of the polynomial function into the corresponding place
		ret.at(i) = param_i;
	}

	return ret;

}





void Function::rotate_vec_func(Function::vec_func* v, double alpha){

	// this function basically uses the theory of a rotation matrix
	// v = D(alpha) * f

	int nx = v->poly_x.size(); // number of parameters in each polynomial function
	int ny = v->poly_y.size();

	// determine the "bigger" polynomial function
	double maxn = ( nx > ny ) ? nx : ny;

	// these vectors store the rotated value for all the parameters
	// by further questions look at the documentation
	// TODO documentation
	vector<double> cosx(maxn);
	vector<double> cosy(maxn);
	vector<double> sinx(maxn);
	vector<double> siny(maxn);

	// now calculate all the needed parameters
	for(int i = 0; i < nx; i++){
		// this part could be split nicely into four different threads depending on speed
		cosx.at(i) = v->poly_x.at(i) * cosinus(alpha);
		sinx.at(i) = v->poly_x.at(i) * sinus(alpha);
	}
	for(int i = 0; i < ny; i++){
		// this part could be split nicely into four different threads depending on speed
		cosy.at(i) = v->poly_y.at(i) * cosinus(alpha);
		siny.at(i) = v->poly_y.at(i) * sinus(alpha);
	}

	// we need as many parameters for the vector function, as the bigger vector function has
	// by initializing over this method, every parameter in beginning we be 0
	// ( this is important later
	v->poly_x = vector<double>(maxn);
	v->poly_y = vector<double>(maxn);

	// this is the step D(alpha) * f
	// the steps before where preparation
	for(int i = 0; i < maxn; i++){
		v->poly_x.at(i) = cosx.at(i) - siny.at(i);
		v->poly_y.at(i) = sinx.at(i) + cosy.at(i);
	}

}



Function::vec_func Function::rotate(vector<double>* poly, double alpha){


	vec_func temp;

	// we just assume x = t for x(t)
	temp.poly_x = {0, 1};
	// and the polynomial function for y(t) is just f(x)
	temp.poly_y = *poly;

	// these are just arbitrarily set, and should be changed accordingly
	temp.end_t = 100;
	temp.start_t = -100;

	// now rotate the temporarily created vector function
	rotate_vec_func(&temp, alpha);
	// and returned it directly after that
	return temp;
}


double Function::find_t(Function::vec_func* v,
						double x,
						double t_step_size,
						double margin){

	// first get a pointer to the polynomial function for x(t)
	vector<double>* xt = &(v->poly_x);

//

	// now save the minimum for the range in
	// which this function is clearly defined
	double t0;
	double t1 = v->start_t;

	// look what value of x is at place t1
	double x_t0;
	double x_t1 = evaluate(xt, t1);

	// if we can't reach it, throw an exception
	if(x < x_t1){
		return -1;
	}

	// now move t_step_size forward on the function
	// as long as x_t1 < x
	while(x_t1 < x){

		// save the old values for the middling part of the algorithm
		x_t0 = x_t1;
		t0 = t1;

		// now move t_step_size forward
		t1 += t_step_size;

		// check if we are still in our well defined range
		if(t1 > v->end_t){

			// if t1 is outside, then see, if x_t0 and v->end_t
			// enclose our x
			t1 = v->end_t;
			x_t1 = evaluate(xt, t1);

			if(x <= x_t1){
				// if x is inside that range,
				// we jump to the middling part of the code
				break;
			}
			// when that is not the case, we throw an exception,
			// since we can't reach x in our well defined range
			else{
				return -1;
			}
		}

		// if we are still in our well defined range, then evaluate x(t) at t1
		x_t1 = evaluate(xt, t1);

	}

	// now start the middling, but for that we need two new values for the
	// middled t and the corresponding x
	double tmid;
	double x_tmid;

	while(true){

		if(abs(x_t0 - x) <= margin){return t0;}
		if(abs(x_t1 - x) <= margin){return t1;}


		// calculate the x-value that is in the middle of t0 and t1
		tmid = (t0 + t1)/2.0;
		// calculate the x, that is corresponding to that
		x_tmid = evaluate(xt, tmid);


		// if x_tmid is bigger than x, then we know, that x is in range [x_t0, x_tmid]
		// so the maximum t value needs to be replaced
		if(x_tmid > x){
			t1 = tmid;
			x_t1 = x_tmid;
		}
		// if x_tmid is smaller than x, then we know, that x is in range [x_tmid, x_t1]
		// so the minimum t value needs to be replaced
		else{
			t0 = tmid;
			x_t0 = x_tmid;
		}

		// this algorithm basically closes in on the searched for t value, step by step
		// and throws out half of the possible remaining answers per step
	}
}

Function::pPoint Function::evaluate_vec_func(Function::vec_func* v, double t){
	// evaluates each polynomial function in turn and then puts them together into a pPoint
	return pPoint(evaluate(&v->poly_x, t), evaluate(&v->poly_y, t));

}

double Function::find_t_for_length(Function::vec_func* v,
								   double start_t,
								   double length,
								   double t_step_size,
								   double margin){

	// this is the distance that has already been traversed.
	double accumulated_distance;
	double distance;

	// these are the t's between small segments are calculated
	double t0;
	double t1 = start_t;

	// first find the point for the start_t
	pPoint pt1;
	pPoint pt2 = evaluate_vec_func(v, t1);

	while(true){

		// this movements of values for the points is setup so, by the end of a
		// while cycle everything is ready to go for the next step of this function
		pt1 = pt2;
		t0 = t1;

		// we go delta steps forward
		t1 += t_step_size;
		// calculate the point for t1
		pt2 = evaluate_vec_func(v, t1);

		// check the direct distance between those points.
		// this is also the reason why decreasing t_step_size increases the accuracy,
		// as there is less error in the approximation

		distance = calc_distance(pt1, pt2);

		if(accumulated_distance + distance >= length){
			break;
		}
		else{
			accumulated_distance += distance;
		}

	}

	double t_mid;

	// as soon as accumulated_distance overshoots the length
	// t gets middled as many times as was stated
	// t0 is always the start point for the calculation
	// (so pt1 stays constant from here on out)
	do{

		// find the point that lies in between the previous points
		t_mid = (t0 + t1)/2;
		pt2 = evaluate_vec_func(v, t_mid);

		distance = calc_distance(pt1, pt2);

		// move the middling range depending on the distance
		// and close in on the actual value
		if(accumulated_distance + distance > length){
			t1 = t_mid;
		}
		else{
			t0 = t_mid;
		}

	} while(abs(accumulated_distance + distance - length) >= margin);

	return t_mid;

}


double Function::calc_length(Function::vec_func* v,
							 double t0,
							 double t1,
							 int guide_points){

	// first calculate how big the delta-t is
	// said t-value will be used to get the continues points between t0 an t1
	double delta_t = (t1 - t0) / (guide_points + 1);

	// then define all the points for which the distance will be measured
	vector<pPoint> points;

	// calculate the first point and then add it to points
	points.push_back(evaluate_vec_func(v, t0));

	for(int i = 0; i < guide_points; i++){
		t0 += delta_t;
		points.push_back(evaluate_vec_func(v, t0));
	}

	// put the last point into points
	points.push_back(evaluate_vec_func(v, t1));

	// now add up all the distances between consecutive points

	double distance = 0;

	int size = points.size();

	for(int i = 0; i < size -1; i++){

		distance += calc_distance(points.at(i), points.at( i+1 ));

	}

	return distance;

}

double Function::calc_distance(pPoint pt1, pPoint pt2){
	double delta_x = pt1.x - pt2.x;
	double delta_y = pt1.y - pt2.y;
	return sqrt( delta_x*delta_x + delta_y*delta_y );
}



double Function::calc_curvature(Function::vec_func* v, double t){
	// first calculate all the needed derivatives
	vector<double> dx = differentiate(&v->poly_x);
	vector<double> ddx = differentiate(&dx);

	vector<double> dy = differentiate(&v->poly_y);
	vector<double> ddy = differentiate(&dy);

	// now calculate all the x and y values at t
	double dx_t = evaluate(&dx, t);
	double ddx_t = evaluate(&ddx, t);
	double dy_t = evaluate(&dy, t);
	double ddy_t = evaluate(&ddy, t);


	// first calculate what is inside the brackets of the denominator
	// to make the calculation after easier to understand in code
	double b = dx_t*dx_t + dy_t*dy_t;

	// now calculate the curvature

	return (dx_t*ddy_t - ddx_t*dy_t) / sqrt(b*b*b);

}


void Function::get_point_representation(Function::vec_func* v,
										vector<Function::pPoint>* out,
										int number_points){

	// first clear the output
	out->clear();

	// is v even a workable function?
	if(!is_usable(v)){
		return;
	}

	*out = vector<Function::pPoint>(number_points);

	// calculate delta t (more or less how far the points will be appart)
	double dt = (v->end_t - v->start_t) / number_points;

	// start at t = start_t
	double t = v->start_t;

	// add number_points points to the output
	for(int i = 0; i < number_points; i++){
		// calculate the point for the vector function and add it to the output
		out->at(i) = evaluate_vec_func(v, t);

		// go delta t steps forward
		t += dt;

	}

}


double Function::calc_theta(Function::vec_func* v, double t){

	// calculate the derivative of the polynomial function
	// and evaluate it at t
	vector<double> dx = differentiate(&v->poly_x);
	double x = evaluate(&dx, t);

	vector<double> dy = differentiate(&v->poly_x);
	double y = evaluate(&dx, t);

	// calculate the the angle
	return arc_tangens(y/x);

}

void Function::calc_middle_vec_func(Function::vec_func* v1,
									Function::vec_func* v2,
									Function::vec_func* out){

	// TODO
	if(!is_usable(v1)){
		return;
	}
	if(!is_usable(v2)){
		return;
	}

	// find the max grad for the polynomial functions
	int v1x_size = v1->poly_x.size();
	int v2x_size = v2->poly_x.size();
	double max_x = ( v1x_size > v2x_size ) ? v1x_size : v2x_size;

	int v1y_size = v1->poly_y.size();
	int v2y_size = v2->poly_y.size();
	double max_y = ( v1y_size > v2y_size ) ? v1y_size : v2y_size;

	// create the poly x and y for the out vector
	out->poly_x = vector<double>(max_x);
	out->poly_y = vector<double>(max_x);

	// go through all the parameters

	//TODO
	// TODO für spätere Versionen muss das hier noch variable für die länge gemacht werden
	// derzeit geht das nur, weil die Polynome definitiv den gleichen Grad haben
	for(int i = 0; i < max_x; i++){
		out->poly_x.at(i) = (v1->poly_x.at(i) + v2->poly_x.at(i))/ 2;
	}

	for(int i = 0; i < max_y; i++){
		out->poly_y.at(i) = (v1->poly_y.at(i) + v2->poly_y.at(i))/ 2;
	}

	out->end_t = (v1->end_t + v2->end_t)/ 2;

	out->start_t = ( v1->start_t > v2->start_t ) ? v1->start_t : v2->start_t;
//	out->start_t = (v1->start_t + v2->start_t)/ 2;


}

bool Function::is_usable(Function::vec_func* v){

	if(v == 0 || // if the reference is not valid
	   v->end_t <= 0 || // is the end value is too small
	   v->poly_x.size() == 0 || // if there is no polynomial function for x
	   v->poly_y.size() == 0){ // if there is no polynomial function for x
		// then v is not workable
		return false;
	}

	// if there happens to be any nan-value inside the polynomial functions
	// then it is also not usable
	int size = v->poly_x.size();
	int i;

	for(i = 0; i < size; i++){
		if(std::isnan(v->poly_x.at(i)))
			return false;
	}

	size = v->poly_y.size();

	for(i = 0; i < size; i++){
		if(std::isnan(v->poly_y.at(i)))
			return false;
	}

	// otherwise it is
	return true;

}

